#
import re

from pdfminer.layout import LTContainer, LTRect, LTLine, LTTextLine

ROUND_LEVEL = 1
HDR_ROW_THRESH = 7


class Element(object):
    def __init__(self, bbox):
        self.bbox = (round(bbox[0], ROUND_LEVEL),
                     round(bbox[1], ROUND_LEVEL),
                     round(bbox[2], ROUND_LEVEL),
                     round(bbox[3], ROUND_LEVEL)
                    )
        self.cent = (round((bbox[0] + bbox[2]) / 2, ROUND_LEVEL),
                     round((bbox[1] + bbox[3]) / 2, ROUND_LEVEL))
        self.area = round(abs((bbox[2] - bbox[0]) * (bbox[3] - bbox[1])),
                          ROUND_LEVEL)
        self.row = None
        self.col = None
        self.orphan = False

class Rect(Element):
    _rectlist = []
    _is_assigned = False
    num_rows = None
    hdr_row = 0

    def __init__(self, bbox):
        Element.__init__(self, bbox)
        Rect._rectlist.append(self)
        Rect._is_assigned = False

    def __unicode__(self):
        return (u"Rect[%s, %s]: @%s (%s)"
                % (self.row, self.col, self.area, self.bbox))

    @classmethod
    def clear(cls):
        cls._rectlist = []
        cls._is_assigned = False

    @classmethod
    def _eliminate_dups(cls):
        cls._rectlist.sort(key=lambda r:r.cent[1])
        cls._rectlist.sort(key=lambda r:r.cent[0])
        dups = []
        for i_, rect in enumerate(cls._rectlist[:-1]):
            if rect.bbox == cls._rectlist[i_+1].bbox:
                dups.append(i_)
        #print ("Rect eliminating %s dups from list of %s."
        #       %(len(dups), len(cls._rectlist)))
        dups.reverse()
        for i_ in dups:
            del cls._rectlist[i_]

    @classmethod
    def find_rect(cls, cent):
        rtn = None
        for rect in cls._rectlist:
            if (    rect.bbox[0] <= cent[0] <= rect.bbox[2] and
                    rect.bbox[1] <= cent[1] <= rect.bbox[3] ):
                if rtn is None:
                    rtn = rect
                else:
                    #print ("OVERLAP: centroid %s in multiple rects: %s, %s"
                    #        % (cent, rtn, rect))
                    # select the rect with the smaller area
                    if rtn.area > rect.area:
                        rtn = rect
        return rtn

    @classmethod
    def _assign_rows(cls):
        # sort by y to find rows
        cls._rectlist.sort(key=lambda r:r.bbox[1])
        span = cls._rectlist[0].bbox[3] - cls._rectlist[-1].bbox[1]
        thresh = abs(span / len(cls._rectlist))
        row = 0
        row_count = 0
        last_y = cls._rectlist[0].bbox[1]
        for rect in cls._rectlist:
            if (rect.bbox[1] - last_y) > thresh:
                if not cls.hdr_row and row_count > HDR_ROW_THRESH:
                    cls.hdr_row = row
                row += 1
                row_count = 0
            rect.row = row
            row_count += 1
            last_y = rect.bbox[1]
        Rect.num_rows = row
        #print ("Rect._assign_rows: %s rows, thresh %s " % (row, thresh))

    @classmethod
    def _assign_cols(cls):
        # sort by x, then by row to find cols
        # this relies on python's stable sort
        cls._rectlist.sort(key=lambda r:r.cent[0])
        cls._rectlist.sort(key=lambda r:r.row)

        col = 0
        last_row = cls._rectlist[0].row
        for rect in cls._rectlist:
            if rect.row != last_row:
                last_row = rect.row
                col = 0
            rect.col = col
            #print "Assigned column: %s" % rect
            col += 1

    @classmethod
    def _check_cols(cls):
        #print "_check_cols: Rect.hdr_row:", cls.hdr_row
        if not cls.hdr_row:
            return
        for rect1 in cls._rectlist:
            if rect1.row != cls.hdr_row: continue
            for rect2 in cls._rectlist:
                if rect2.row <= cls.hdr_row: continue
                if ( abs(rect1.bbox[0] - rect2.bbox[0])
                      + abs(rect1.bbox[2] - rect2.bbox[2])
                      < 0.01 * rect1.bbox[0] ) :
                    if rect2.col < rect1.col:
                        #print "\t1:", rect2, "to", rect1
                        rect2.col = rect1.col
                elif (rect2.orphan and rect2.col < rect1.col
                        and rect2.bbox[0] >= rect1.bbox[0]
                        and rect2.bbox[2] <= rect1.bbox[2]):
                    #print "\t2:", rect2, "to", rect1
                    rect2.col = rect1.col


    @classmethod
    def assign_locations(cls):
        if cls._is_assigned:
            return
        cls._eliminate_dups()
        cls._assign_rows()
        cls._assign_cols()
        cls._check_cols()
        cls._is_assigned = True

    @classmethod
    def get_location(cls, cent):
        cls.assign_locations()
        rect = cls.find_rect(cent)
        if rect is None:
            return None, None
        return rect.row, rect.col


class Text(Element):
    _textlist = []
    _is_assigned = False

    def __init__(self, bbox, text):
        Element.__init__(self, bbox)
        self.text = text
        self.rect = None
        Text._textlist.append(self)
        Text._is_assigned = False

    def __unicode__(self):
        return (u"Text[%s, %s]: (%s) '%s'"
                % (self.row, self.col, self.cent, self.text))

    @classmethod
    def clear(cls):
        cls._textlist = []
        cls._is_assigned = False

    @classmethod
    def assign_locations(cls):
        if cls._is_assigned:
            return

        # assign each Text to a Rect
        # this can introduce new Rect and Text objects
        cls._assign_rect()

        # assign locations to text
        Rect.assign_locations() # assign a row/col to each rect
        for text in cls._textlist:
            text.row, text.col = Rect.get_location(text.cent)
            #print text

    @classmethod
    def _assign_rect(cls):
        # find orphans and create Rect objects from their bboxs
        for text in cls._textlist:
            text.rect = Rect.find_rect(text.cent)

            if text.rect is None:
                #print "Orphan text:", text
                text.rect = Rect(text.bbox)
                text.orphan = True
                text.rect.orphan = True

        # find text spanning rects
        # This occurs when adjacent cells are right and left justified
        # respectively.  That means a number followed  by text.
        # Make separare text item of leading number.
        split_text_objects = []
        for i_, text in enumerate(cls._textlist):
            if (text.bbox[0] >= (text.rect.bbox[0] -2) and
                text.bbox[2] <= (text.rect.bbox[2] +2) ):
                continue
            #Any text not passing the above test spans multiple rects.

            # If we can split the text and find Rects for the two parts
            # We will create two new Text objects and delete the old one
            #print "Splitting:", text
            fl = re.split(r'([-+]?[0-9]*\.?[0-9]+)', text.text, 1)
            if len(fl) == 3:
                text0 = fl[1]
                text1 = fl[2]
                cent0 = text.bbox[0], text.cent[1]
                cent1 = text.bbox[2], text.cent[1]
                rect0 = Rect.find_rect(cent0)
                rect1 = Rect.find_rect(cent1)
                if rect0 and rect1:
                    #print "\tSplit '%s' & '%s'."%(text0, text1)
                    t0 = Text((text.bbox[0], text.bbox[1],
                               rect0.bbox[2], rect0.bbox[3]),
                              text0)
                    t0.rect = rect0
                    t1 = Text((rect1.bbox[0], rect1.bbox[1],
                               text.bbox[2], text.bbox[3]),
                              text1)
                    t1.rect = rect1
                    split_text_objects.append(i_)
        split_text_objects.reverse()
        for i_ in split_text_objects: del cls._textlist[i_]


    @classmethod
    def get_row_text(cls, row):
        rowlist = [t for t in cls._textlist if t.row == row]
        rowlist.sort(key=lambda t:t.cent[1])
        rowlist.sort(key=lambda t:t.col)
        #print 'rowlist %s: %s'%row, rowlist
        col = 0
        coltext = ['']
        rowtext = []
        for text in rowlist:
            while text.col > col:
                rowtext.append(coltext)
                coltext = ['']
                col += 1
            coltext.append(text.text)
        rowtext.append(coltext)
        #print "Rowtext %s:"%row, rowtext
        return rowtext


class Sheet2 (object):
    def __init__(self):
        Rect.clear()
        Text.clear()

    def add_ltcontainer (self, obj, page_y_offset):
        #NB: row indexes (y axis) are negative!
        bbox = (obj.x0, -(obj.y1 + page_y_offset),
                obj.x1, -(obj.y0 + page_y_offset))

        if isinstance (obj, LTTextLine):
            txt = unicode(obj.get_text().strip())
            #print \
            Text(bbox, txt)
        elif isinstance (obj, LTRect):
            #print \
            Rect(bbox)
        elif isinstance (obj, LTContainer):
            for child in obj:
                self.add_ltcontainer (child, page_y_offset)
        else:
            pass

    def extract_rows (self):
        Text.assign_locations()
        rows = []
        for row in range(Rect.num_rows):
            rowtext = Text.get_row_text(row)
            joined_cols = [' '.join(c).strip() for c in rowtext]
            rows.append(joined_cols)
        return rows




#

from pdfminer.layout import LTContainer, LTComponent, LTRect, LTLine, LAParams, LTTextLine
from pdfminer.utils import Plane


class Logger (object):

    def __init__(self):
        self.messages = []

    def log(self, severity, message):
        self.messages.append({'severity':severity, 'message': message})

    def warning (self, message):
        self.log('warning', message)

    def error (self, message):
        self.log('error', message)

    def has_error (self):
        for m in self.messages:
            if m['severity'] == 'error': return True
        return False

    def has_warning (self):
        for m in self.messages:
            if m['severity'] == 'warning': return True
        return False

    def has_message (self, severity, message):
        for m in self.messages:
            if m['severity'] == severity and m['message'] == message : return True
        return False

    def get_messages (self):
        return '\n'.join(['%(severity)s: %(message)s' % (m) for m in self.messages])


class CellText (LTComponent):
    text = ''

    def __init__(self, bbox, text):
        LTComponent.__init__ (self, bbox)
        self.text = text

class Cell (LTComponent):
    text_lines = None

    def __init__(self, bbox):
        LTComponent.__init__ (self, bbox)
        self.text_lines = []

    def get_text (self):
        return ' '.join ([t.text for t in self.text_lines])

class Sheet1 (object):
    cells = None
    text_layout = None
    column_edges = None
    row_edges = None


    def __init__(self):
        self.cells = Plane()
        self.text_layout = Plane()
        self.row_edges = {}
        self.column_edges = {}

    def add_cell (self, cell):
        self.cells.add(cell)


    def add_text (self, cell_text):
        self.text_layout.add(cell_text)
#        if cell_text.text[:3] == 'Oil': print cell_text.text, cell_text.bbox

    def add_column_edge (self, x_value):
        x = round(x_value,2)
        self.column_edges[x] = 1+ self.column_edges.get(x,0)

    def add_row_edge (self, y_value):
        y = round(y_value,2)
        self.row_edges[y] = 1+ self.row_edges.get(y,0)

    def add_line (self, bbox):
        if bbox[0]==bbox[2]:
            # vertical line
            self.add_column_edge(bbox[0])
        elif bbox[1]==bbox[3]:
            #horizontal line
            self.add_row_edge(bbox[1])
        else:
            print ('WARNING: non-orthogonal line found: %s'%bbox)

    def add_rect (self, bbox):
        self.add_column_edge(bbox[0])
        self.add_column_edge(bbox[2])
        self.add_row_edge(bbox[1])
        self.add_row_edge(bbox[3])

    def add_ltcontainer (self, obj, page_y_offset):
        #NB: row indexes (y axis) are negative!

        bbox = (
            round(obj.x0,2),
            round(-(obj.y1+page_y_offset),2),
            round(obj.x1,2),
            round(-(obj.y0+page_y_offset),2)
            )

        if isinstance (obj, LTTextLine):
            self.add_text (CellText(bbox, obj.get_text()))
        elif isinstance (obj, LTLine):
            self.add_line(bbox)
        elif isinstance (obj, LTRect):
            self.add_rect(bbox)
        elif isinstance (obj, LTContainer):
            for child in obj:
                self.add_ltcontainer (child, page_y_offset)
        else:
            pass

    def extract_rows (self):
#        for obj in self.text_layout.find((690, -1200, 800, -1000)):
#            print obj.bbox,obj.text

        row_bounds = sorted(self.row_edges)
        col_bounds = sorted(self.column_edges)

#        pprint.pprint(col_bounds)

        rows = []

        r0 = row_bounds[0] - 1 if row_bounds else 0

        #NB: row indexes (y axis) are negative!
        for r1 in row_bounds:
            if r1 - r0 < 1: continue

#            print r1-r0,r0,r1

            row=[]
            c0 = 0
            for c1 in col_bounds:
                if c1 - c0 < 1: continue

#                print c0,r0,c1,r1
                # get all text lines that intersect the bounds of this cell
                lines = [l for l in self.text_layout.find((c0,r0,c1,r1))]
                #sort from top to bottom
                lines = sorted(lines, key=lambda line: line.y0)

#                text = ' '.join([t.text.strip() for t in lines if t.x0 >= c0 and t.x0 <= c1])
#                if text[:10] == 'Production': print text,c0,r0,c1,r1
#                if text[:3] == 'Oil': print text,c0,r0,c1,r1

                # remove anything where the left edge is not inside the cell and concatenate the rest
                row.append(' '.join([t.text.strip() for t in lines if t.x0 >= c0 and t.x0 <= c1]))
                c0 = c1
            rows.append(row)
            r0 = r1
        return rows



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='fracfocustools',
    version='0.0.5',
    description='Tools for maipulating data acquired from fracfous.org',
    long_description=readme,
    author='Paul Woods, SkyTruth',
    author_email='paul@skytruth.org',
    url='https://github.com/SkyTruth/fracfocustools',
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=["pdfminer"],
    package_data = {'': ['README.md', 'LICENSE']},
    include_package_data = True
)



# -*- coding: utf-8 -*-

from StringIO import StringIO
import sys, traceback
import re


from pdfminer.pdfparser import PDFParser, PDFDocument
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfdevice import PDFDevice
from pdfminer.layout import LTContainer, LTComponent, LTRect, LTLine, LAParams, LTTextLine
from pdfminer.utils import Plane
from pdfminer.converter import PDFPageAggregator

from helpers1 import Sheet1
from helpers2 import Sheet2

class Report (object):
    field_defs = None
    report_data = None
    col_indexes = None
    messages = None
    chemicals = None

    def __init__ (self, logger):
        self.field_defs = [
            {'label':'fracture_date', 'regex': 'fracturedate:?$', 'type': 'report'},
            {'label':'fracture_date', 'regex': 'lastfracturedate:?$', 'type': 'report'},
            {'label':'fracture_date', 'regex': 'jobstartdate:?$', 'type': 'report'},
            {'label':'state', 'regex': 'state:?$', 'type': 'report'},
            {'label':'county', 'regex': 'county:?$', 'type': 'report'},
            {'label':'county', 'regex': 'county/parish:?$', 'type': 'report'},
            {'label':'api', 'regex': 'apinumber:?$', 'type': 'report'},
            {'label':'operator', 'regex': 'operatorname:?$', 'type': 'report'},
            {'label':'well_name', 'regex': 'wellnameandnumber:?$', 'type': 'report'},
            {'label':'longitude', 'regex': 'longitude:?$', 'type': 'report'},
            {'label':'latitude', 'regex': 'latitude:?$', 'type': 'report'},
            {'label':'datum', 'regex': 'long/latprojection:?$', 'type': 'report'},
            {'label':'datum', 'regex': 'datum:?$', 'type': 'report'},
            {'label':'total_water_volume', 'regex': 'totalwatervolume.*$', 'type': 'report', 'optional': True},
            {'label':'total_water_volume', 'regex': 'totalbasewatervolume.*$', 'type': 'report', 'optional': True},

            # Only in old format report fields
            {'label':'production_type', 'regex': 'productiontype:?$', 'type': 'report', 'optional': True},
            {'label':'true_vertical_depth', 'regex': 'trueverticaldepth.*$', 'type': 'report', 'optional': True},
            # Only in new format report fields
            {'label':'federal', 'regex': 'federal.*$', 'type': 'report', 'optional': True},
            {'label':'base_nonwater_volume', 'regex': 'totalbasenonwatervolume.*$', 'type': 'report', 'optional': True},
            {'label':'end_date', 'regex': 'jobenddate:?$', 'type': 'report', 'optional': True},

            {'label':'trade_name', 'regex': 'tradename$', 'type': 'column'},
            {'label':'supplier', 'regex': 'supplier$', 'type': 'column'},
            {'label':'purpose', 'regex': 'purpose$', 'type': 'column'},
            {'label':'ingredients', 'regex': 'ingredients$', 'type': 'column'},
            {'label':'cas_number', 'regex': 'chemicalabstractservicenumber', 'type': 'column'},
            {'label':'additive_concentration', 'regex': 'maximumingredientconcentrationinadditive', 'type': 'column'},
            {'label':'hf_fluid_concentration', 'regex': 'maximumingredientconcentrationinhffluid', 'type': 'column', 'optional':True},
            {'label':'ingredient_weight', 'regex': 'maximumingredientweight', 'type': 'column', 'optional': True},
            {'label':'comments', 'regex': 'comments$', 'type': 'column'},

            {'label':'footer', 'regex': '\*totalwatervolumesourcesmayinclude', 'type': 'footer'},
            {'label':'footer', 'regex': '\*totalchemicalmassisthetotalamountof', 'type': 'footer'},

            {'label':'ignore', 'regex': 'ingredientsshownabove', 'type': 'divider', 'optional': True},
        ]
        self.global_filter = '\(cid:.*?\)'  # infrequently occurring extraneous data in some pdfs
        self.report_data = {}
        self.col_indexes = {}
        self.chemicals = []
        self.logger = logger

    def __unicode__(self):
        rptlist = []

        rpt_items = self.report_data.items()
        rpt_items.sort(key=lambda i:i[0])
        rptlist.extend([u'\t'.join((str(item[0]), str(item[1])))
                for item in rpt_items])
        hdr_items = self.col_indexes.items()
        hdr_items.sort(key=lambda i:i[1])
        rptlist.append(u'\t'.join([item[0] for item in hdr_items]))
        self.col_indexes['row'] = -1
        for chemical in self.chemicals:
            chm_items = chemical.items()
            chm_items.sort(key=lambda i:self.col_indexes[i[0]])
            rptlist.append(u'\t'.join([item[1] for item in chm_items[1:]]))
        del self.col_indexes['row']
        rpt = u'\n'.join([unicode(l) for l in rptlist])
        assert isinstance(rpt, unicode)
        return rpt

    def __str__(self):
        return self.__unicode__().encode('ascii', 'replace')

    def warning (self, msg):
        self.logger.warning(msg)

    def error (self, msg):
        self.logger.error(msg)

    def filter_field(self, field):
        return re.sub(self.global_filter, '', field)

    def match_field_def (self, text, field_def):
        t = re.sub('[\s]','',text)
        return re.match(field_def['regex'], t, re.IGNORECASE)

    def find_field_def (self, text):
        t = re.sub('[\s]','',text)
        for d in self.field_defs:
            if re.match(d['regex'], t, re.IGNORECASE):
                return d
        return None

    def row_contains_label (self, row):
        for col in row:
            field_def = self.find_field_def (col)
            if field_def: return field_def
        return None

    def extract_report_field (self, field_def, row):
        text= []

        for col in row:
#            print field_def['label'],col
            # skip data that matches the field label
            if col and not self.match_field_def(col,field_def):
                #take only the first non-blank column that does not match the field label
                if col:
                    text.append (col)
                    break
        filtered_text = self.filter_field(' '.join(text))
        self.report_data [field_def['label']] = filtered_text

    def extract_column_headings (self, row):
        for i,text in enumerate(row):
            if text:
                field_def = self.find_field_def(text)
                if field_def and field_def['type'] == 'column':
                    self.col_indexes[field_def['label']] = i
                else:
                    self.warning('Unknown column heading: %s'%text)

        # check to see that we got all the coluumn headings
        expected_count = 0
        for d in self.field_defs:
            if d['type'] == 'column' and not d.get('optional'):
                expected_count += 1
                if d['label'] not in self.col_indexes:
                    self.error('Missing column header: %s' %d['label'])


        if len(self.col_indexes) < expected_count:
            self.error('Wrong columnn heading count - expected: %s, found: %s'%(expected_count, len(self.col_indexes)))

    def extract_chemical_record (self, row, chem_row_number):
        record = {'row': chem_row_number}
        for label,i in self.col_indexes.items():
            try:
                record[label] = self.filter_field(row[i])
            except IndexError:
                record[label] = ''

        self.chemicals.append(record)

    def extract_data (self, sheet):
        rows = sheet.extract_rows()

        found_col_header = False
        chem_row_number = 1

        for row in rows:
            #print "ROW:", row
            field_def = self.row_contains_label (row)
            if field_def and field_def['type'] == 'ignore': continue
            if found_col_header:
                # if we have already seen the column header row, extract a chemical record and add to the list of chemicals

                if field_def and field_def['type'] == 'footer':
                    # found the footer - no more data after this
                    break

                if field_def and field_def['type'] == 'divider':
                    record = {'row': chem_row_number, 'trade_name':''.join(row)}
                    self.chemicals.append(record)
                    continue
                elif ''.join(row):
                    # don't extract empty rows
                    self.extract_chemical_record (row, chem_row_number)
                chem_row_number += 1
            elif field_def:
                # see if this row contains a defined label
                # if this is a report row, extract and add to the report
                if field_def['type'] == 'report':
                    self.extract_report_field (field_def, row)
                elif field_def['type'] == 'column':
                    # if this is the column header row, extract and set column indexes
                    self.extract_column_headings(row)
                    found_col_header = True
            else:
#                print row
                pass

        for d in self.field_defs:
            if d['type'] == 'report' and d['label'] not in self.report_data:
                if d.get('optional'):
                    self.warning('Missing report field: %s' %d['label'])
                else:
                    self.error('Missing report field: %s' %d['label'])


class FracFocusPDFParser (object):

    def __init__ (self, raw_pdf, logger):
        self.raw_pdf = raw_pdf
        self.logger = logger
        self.report = None

    def find_pdf_text(self, obj, text):
        if isinstance (obj, LTTextLine):
            if obj.get_text().strip() == text:
                return True
        elif isinstance (obj, LTContainer):
            for child in obj:
                if self.find_pdf_text(child, text):
                    return True
        return False

    def parse_pdf (self):
        self.report = Report (self.logger)
        fp = StringIO(self.raw_pdf)
        parser = PDFParser(fp)
        doc = PDFDocument()
        parser.set_document(doc)
        try:
            doc.set_parser(parser)
            doc.initialize('')
            if not doc.is_extractable:
                raise RuntimeError("PDFTextExtractionNotAllowed")

            rsrcmgr = PDFResourceManager()
            laparams = LAParams(
                                char_margin=0.01,       # default 1.0
                                word_margin=0.2,       # default 0.2
                                line_margin=0.3,       # default 0.3
                                line_overlap=0.5       # default 0.5
                               )
            device = PDFPageAggregator(rsrcmgr, laparams=laparams)
            interpreter = PDFPageInterpreter(rsrcmgr, device)

            page_y_offset = 0
            pages = []
            for page in doc.get_pages():
                pages.append(page)
            pages.reverse()

            # Detect pdf format
            for page in pages:
                interpreter.process_page(page)
                layout = device.get_result()
                if self.find_pdf_text(layout, "Job Start Date:"):
                    sheet = Sheet2()
                    break
            else:
                sheet = Sheet1()
                #sheet = Sheet2()  #  TESTING

            for page in pages:
                interpreter.process_page(page)
                layout = device.get_result()
                sheet.add_ltcontainer (layout, page_y_offset)
                page_y_offset += layout.y1

            self.report.extract_data (sheet)
        except Exception:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            trace = traceback.format_exception (exc_type, exc_value, exc_traceback)
            self.logger.error('%s'%''.join(trace))

        if self.logger.has_error():
            return None
        else:
            return self.report